from dataclasses import dataclass
from typing import Optional


class Node:
    """Represents a node in the abstract syntax tree"""
    @property
    def precedence(self):
        return 0
    original = None
    # TODO: If we typehint this as "Node", get_type_hints(ast.Ellipsis()), get_type_hints(ast.Break()) and
    # get_type_hints(ast.Nil()) crash in optimize.apply_trans, because it uses type hints to figure out
    # which children should the transformation be applied progressively. At the moment, it's not an issue
    # if the apply_trans does not get applied to "original" attributes, as it is currently used only to
    # transform loads to funcs, and at that point, there should be no originals.


@dataclass
class Name(Node):
    """Represents an variable identifier in the abstract syntax tree"""
    id: str


@dataclass
class Block(Node):
    """Represents list of statements in the abstract syntax tree"""
    stats: list[Node]


@dataclass
class Return(Node):
    """Represents a return statement in the abstract syntax tree"""
    exps: list[Node]


@dataclass
class Do(Node):
    """Represents a 'do...end' in the abstract syntax tree"""
    block: Block


@dataclass
class Assign(Node):
    """Represents an assignment in the abstract syntax tree"""
    targets: list[Node]
    values: list[Node]


@dataclass
class Label(Node):
    """Represents a '::label::' in the abstract syntax tree"""
    name: str


@dataclass
class Break(Node):
    """Represents a 'break' in the abstract syntax tree"""
    pass


@dataclass
class LiteralString(Node):
    """Represents a string literal in the abstract syntax tree"""
    value: str


@dataclass
class Goto(Node):
    """Represents a 'goto target' in the abstract syntax tree"""
    target: str


@dataclass
class Boolean(Node):
    """Represents a boolean literal in the abstract syntax tree"""
    value: bool


@dataclass
class Ellipsis(Node):
    """Represents an ellipsis (...) in the abstract syntax tree"""
    pass


@dataclass
class Nil(Node):
    """Represents a nil literal the abstract syntax tree"""
    pass


@dataclass
class While(Node):
    """Represents a while loop in the abstract syntax tree"""
    condition: Node
    block: Block


@dataclass
class Repeat(Node):
    """Represents a repeat-until loop in the abstract syntax tree"""
    condition: Node
    block: Block


@dataclass
class ForRange(Node):
    """Represents a ranged for loop in the abstract syntax tree"""
    var: Node
    lb: Node
    ub: Node
    step: Optional[Node]  # can be empty
    body: Block


@dataclass
class ForIn(Node):
    """Represents a for-in loop in the abstract syntax tree"""
    names: list[Name]
    exps: list[Node]
    body: Block


@dataclass
class Local(Node):
    """Represents a local assignment in the abstract syntax tree"""
    targets: list[Name]
    values: Optional[list[Node]]


@dataclass
class Func(Node):
    """Represents a function definition the abstract syntax tree"""
    args: list[Name]
    body: Block
    oneline: bool = True


@dataclass
class Index(Node):
    """Represents an indexing (obj[item]) the abstract syntax tree"""
    obj: Node
    item: Node


@dataclass
class Table(Node):
    """Represents a table definition the abstract syntax tree"""
    fields: list[Node]


@dataclass
class Field(Node):
    """Represents a field in the inside a table definition"""
    value: Node
    key: Node = None


@dataclass
class Call(Node):
    """Represents a call in the abstract syntax tree"""
    func: Node
    args: list[Node]


@dataclass
class MethodCall(Node):
    """Represents a method call ('obj:method(args)') in the abstract syntax tree"""
    value: Node
    method: str
    args: list[Node]


@dataclass
class If(Node):
    """Represents an if statement in the abstract syntax tree"""
    test: Node
    body: Block
    orelse: Block


@dataclass
class BinOp(Node):
    """Represents a binary operator in the abstract syntax tree"""
    left: Node
    op: str
    right: Node

    @property
    def precedence(self):
        return _precedence[self.op]


@dataclass
class Hint(Node):
    """A node containing hints for printing. Does not reflect actual code."""
    block: Block
    no_hex: bool = True
    double_quotes: bool = False


_precedence = \
    dict.fromkeys(['--|'], 1) | \
    dict.fromkeys(['^'], 2) | \
    dict.fromkeys(['*', '/', '//', '%'], 3) | \
    dict.fromkeys(['+', '-'], 4) | \
    dict.fromkeys(['..'], 5) | \
    dict.fromkeys(['<<', '>>'], 6) | \
    dict.fromkeys(['&'], 7) | \
    dict.fromkeys(['~'], 8) | \
    dict.fromkeys(['|'], 9) | \
    dict.fromkeys(['<', '>', '<=', '>=', '~=', '=='], 10) | \
    dict.fromkeys(['and'], 11) | \
    dict.fromkeys(['or'], 12)


@dataclass
class UnaryOp(Node):
    """Represents an unary operator in the abstract syntax tree"""
    op: str
    operand: Node

    @property
    def precedence(self):
        return 2


@dataclass(init=False)
class Numeral(Node):
    """Represents a numeral literal in the abstract syntax tree"""
    whole: int
    fractional: int
    exponent: int
    hex: bool

    def __init__(self, whole=0, fractional=0, exponent=0, hex=False):
        self.whole = whole
        self.fractional = fractional
        self.exponent = exponent
        self.hex = hex

    def __repr__(self):
        if self.whole == 0 and self.fractional > 0:
            return f"{'0x.' + ('%x' % self.fractional)[::-1]}{'p' + str(self.exponent) if self.exponent != 0 else ''}" if self.hex \
                else f"{'.' + str(self.fractional)[::-1]}{'e' + str(self.exponent) if self.exponent != 0 else ''}"
        else:
            return f"0x{'%x' % self.whole}{'.' + ('%x' % self.fractional)[::-1] if self.fractional != 0 else ''}{'p' + str(self.exponent) if self.exponent != 0 else ''}" if self.hex \
                else f"{self.whole}{'.' + str(self.fractional)[::-1] if self.fractional != 0 else ''}{'e' + str(self.exponent) if self.exponent != 0 else ''}"

    @property
    def value(self):
        if self.fractional == 0 and self.exponent == 0:
            return float(self.whole)
        else:
            r = repr(self)
            return float.fromhex(r) if self.hex else float(r)


@dataclass
class Alt(Node):
    """
    Represents a list of alternative expressions in the abstract syntax tree.
    Not part of standard LUA, but created using pakettic magic comments.
    Only the first alternative is the one currently active.
    """
    alts: list[Node]

    @property
    def precedence(self):
        return self.alts[0].precedence


@dataclass
class Perm(Node):
    """
    Represents a list of statements that can be freely reordered in the abstract syntax tree.
    Not part of standard LUA, but created using pakettic magic comments.
    """
    stats: list[Node]
    allow_reorder: bool = True



import io
from lib2to3.pgen2.token import SLASH
import os
import struct
import pyparsing as pp
from typing import ByteString
from enum import IntEnum
import typing


class ChunkID(IntEnum):
    """
    Enumerates different chunk types in a TIC-80 Cart
    See: https://github.com/nesbox/TIC-80/wiki/.tic-File-Format
    """
    TILES = 1          # 8 banks, copied to RAM at 0x4000...0x5FFF
    SPRITES = 2        # 8 banks, copied to RAM at 0x6000...0x7FFF
    MAP = 4            # 8 banks, copied to RAM at 0x8000...0xFF7F
    CODE = 5           # 8 banks (PRO only), in ASCII text format
    FLAGS = 6          # 8 banks, sprite flags. copied to RAM at 0x14404...0x14603
    SAMPLES = 9        # 8 banks, copied to RAM at 0x100E4...0x11163
    WAVEFORM = 10      # copied to RAM at 0x0FFE4...0x100E3
    PALETTE = 12       # copied to RAM at 0x3FC0...0x3FEF
    MUSIC = 14         # 8 banks, copied to RAM at 0x13E64...0x13FFB
    PATTERNS = 15      # 8 banks, copied to RAM at 0x11164...0x13E63
    DEFAULT = 17       # flag, no actual content. cart should load default chunk first.
    SCREEN = 18        # 8 banks, 240 x 136 x 4bpp raw buffer. Bank 0 is the cover image.
    BINARY = 19        # 4 banks (?), store binary WASM files.
    COVER_DEP = 3      # deprecated as of 0.90
    PATTERNS_DEP = 13  # 8 banks, deprecated as of 0.80 (copied to RAM at 0x11164...0x13E63)
    CODE_ZIP = 16      # compressed with ZLIB. deprecated as of 1.00, won't be removed.


DATACHUNK_ADDRESSES = {
    ChunkID.TILES: 0x4000,
    ChunkID.SPRITES: 0x6000,
    ChunkID.MAP: 0x8000,
    ChunkID.FLAGS: 0x14404,
    ChunkID.SAMPLES: 0x100E4,
    ChunkID.WAVEFORM: 0x0FFE4,
    ChunkID.PALETTE: 0x3FC0,
    ChunkID.MUSIC: 0x13E64,
    ChunkID.PATTERNS: 0x11164
}

# Each cart may have 1-8 banks, which may or may not be present
Cart = dict[tuple[int, ChunkID], ByteString]


def write_tic(cart: Cart, file: typing.BinaryIO, pedantic=False) -> int:
    """
    Write a .tic cart to a file
        Parameters:
            cart (dict): A dictionary with (bank,chunkId) keys and ByteStrings values
            file (BinaryIO): File to which write the cart
        Returns:
            filesize (int): Size of the just-written cart on disk
    """
    total_size = 0
    for i, bank_chunk in enumerate(cart):
        # if this the last chunk and it's a DEFAULT chunk
        bank, chunk = bank_chunk
        if not pedantic and chunk == ChunkID.DEFAULT and i == len(cart) - 1:
            total_size += file.write(struct.pack("<B", ChunkID.DEFAULT))
            break
        packed_bank_chunk = (bank << 5) + (chunk & 31)
        data = cart[bank_chunk].rstrip(b'\0') if chunk != ChunkID.CODE_ZIP and chunk != ChunkID.CODE else cart[bank_chunk]
        if chunk != ChunkID.DEFAULT and len(data) == 0:
            continue
        size = len(data)
        if (chunk == ChunkID.CODE or chunk == ChunkID.BINARY):
            if size == 65536:
                size = 0
            elif size > 65536:
                raise Exception(f"{chunk.name} chunk in bank {bank} is {size} bytes, maximum size 65536 bytes")
        else:
            if size > 65535:
                raise Exception(f"{chunk.name} chunk in bank {bank} is {size} bytes, maximum size 65535 bytes")
        header = struct.pack("<BHB", packed_bank_chunk, size, 0)  # the last byte is reserved
        total_size += file.write(header)
        total_size += file.write(data)
    return total_size


def read_tic(file: typing.BinaryIO) -> Cart:
    """
    Read a .tic cart from a file
        Parameters:
            file (typing.BinaryIO): File from which to load the cart
        Returns:
            cart (dict): A dictionary with (bank,chunkId) keys and ByteStrings values
    """
    cart = {}
    while True:
        bank_and_chunk_bytes = file.read(1)  # byte, with 3 highest bits bank and 5 lowest bits chunk type
        if len(bank_and_chunk_bytes) == 0:
            return cart
        bank = bank_and_chunk_bytes[0] >> 5
        chunk_id = ChunkID(bank_and_chunk_bytes[0] & 31)
        header = file.read(3)  # sizecoding hackers usually end CHUNK_DEFAULT abruptly without header
        size = struct.unpack("<H", header[:2])[0] if len(header) > 2 else 0
        if (chunk_id == ChunkID.CODE or chunk_id == ChunkID.BINARY) and size == 0:
            size = 65536
        chunk = file.read(size)
        cart[bank, chunk_id] = chunk
        if len(chunk) < size:
            return cart


def split_code(cart: Cart) -> Cart:
    """
    Split the code chunk into at most 8 chunks of 65536 bytes each
    Assumes code is all in bank 0
    Keeps the rest of chunk ordering intact
    """
    codecount = len(list(_ for k, _ in cart.items() if k[1] == ChunkID.CODE))
    if codecount > 1:
        raise Exception("Cannot split code chunk if there are multiple code chunks")
    if codecount == 0:
        return cart
    code = cart[0, ChunkID.CODE]
    if len(code) <= 65536:
        return cart
    ret = {}
    for k, v in cart.items():
        if k[1] != ChunkID.CODE:
            ret[k] = v
            continue
        for i in range((len(code) - 1) // 65536, -1, -1):
            if len(code) <= 65536:
                ret[i, ChunkID.CODE] = code
                break
            else:
                ret[i, ChunkID.CODE] = code[:65536]
                code = code[65536:]
    return ret


def join_code(cart: Cart) -> Cart:
    """
    Join the code chunks into a single chunk
    Places the joined code at the where first code chunk was
    Keeps the rest of chunk ordering intact
    """
    if len(list(_ for k, _ in cart.items() if k[1] == ChunkID.CODE)) <= 1:
        return cart
    code = b''
    for i in range(7, -1, -1):
        if (i, ChunkID.CODE) in cart:
            code += cart[i, ChunkID.CODE]
    ret = {}
    for k, v in cart.items():
        if k[1] != ChunkID.CODE:
            ret[k] = v
            continue
        if code:
            ret[0, ChunkID.CODE] = code
            code = b''
    return ret


def write_lua(cart: Cart, file: typing.TextIO) -> int:
    """
    Write a .lua cart to a file
        Parameters:
            cart (dict): A dictionary with (bank,chunkId) keys and ByteStrings values
            file (TextIO): File to which write the cart
        Returns:
            filesize (int): Size of the just-written cart on disk
    """
    total_size = 0
    total_size += file.write(cart[0, ChunkID.CODE].decode("ascii"))
    for k, v in _TEXTCHUNKS.items():
        num, size, flip, chunk_id = v
        if (0, chunk_id) in cart:
            total_size += file.write(f'\n-- <{k}>')
            for i in range(num):
                part = cart[0, chunk_id][i * size:(i + 1) * size]
                if all(x == 0 for x in part):
                    continue
                hex = part.hex()
                if flip:
                    hex = ''.join([c[1] + c[0] for c in zip(hex[::2], hex[1::2])])
                total_size += file.write(f'\n-- {i:03d}:{hex}')
            total_size += file.write(f'\n-- </{k}>\n')
    return total_size


def read_lua(file: typing.TextIO) -> Cart:
    """
    Read a .lua cart from a file
        Parameters:
            file (TextIO): File from which to load the cart
        Returns:
            cart (dict): A dictionary with (bank,chunkId) keys and ByteStrings values
    """
    code = file.read()
    ret_code = code
    ret = Cart()
    for t, s, e in reversed(list(_tag.scanString(code))):
        num, size, flip, chunk_id = _TEXTCHUNKS[t[0]]
        defined = {int(x[0]): ''.join([c[1] + c[0] for c in zip(x[1][::2], x[1][1::2])]) if flip else x[1] for x in t[1]}
        chunk = b''.join((bytes.fromhex(defined[i]).ljust(size, b'\0')
                         if i in defined else bytes(size) for i in range(num)))
        ret[0, chunk_id] = chunk
        ret_code = ret_code[:s] + ret_code[e:]
    ret_code = ret_code.rstrip()
    ret[0, ChunkID.CODE] = ret_code.encode("ascii")
    return ret


def read(filepath: str) -> Cart:
    """
    Read a tic cart from the disk, detecting the file type based on extension
        Parameters:
            filepath (str): Path of the cart on disk
        Returns:
            cart (dict): A dictionary with (bank,chunkId) keys and ByteStrings values
    """
    _, ext = os.path.splitext(filepath)
    if ext == '.lua':
        with io.open(filepath, "r") as file:
            return read_lua(file)
    else:
        with io.open(filepath, "rb") as file:
            return read_tic(file)


def write(cart: Cart, filepath: str, ext: str = None, pedantic=False) -> int:
    """
    Write a tic cart to the disk
        Parameters:
            cart (dict): A dictionary with (bank,chunkId) keys and ByteStrings values
            filepath (str): Path of the cart on disk
            ext (str): Optional file format extension. If omitted, the format is auto-detected from the filepath.
        Returns:
            filesize (int): Size of the just-written cart on disk
    """
    if ext is None:
        _, ext = os.path.splitext(filepath)
    if ext == '.lua':
        with io.open(filepath, "w", newline='\n') as file:
            return write_lua(cart, file)
    else:
        with io.open(filepath, "wb") as file:
            return write_tic(cart, file, pedantic=pedantic)


def data_to_code(cart: Cart):
    codechunks = [c for c in cart if c[1] == ChunkID.CODE]
    for c in codechunks:
        code = cart[c].decode("ascii")
        bank = c[0]
        noncodechunks = [c for c in cart if c[0] == bank and c[1] in DATACHUNK_ADDRESSES]
        for e in noncodechunks:
            addr = DATACHUNK_ADDRESSES[e[1]]
            datastr = cart[e].hex()
            loader = "i=" + str(addr) + "\nfor m in string.gmatch('" + datastr + "', '%x%x') do\n  poke(i,tonumber(m,16))\n  i=i+1\nend\n"
            code = loader + code
            del cart[e]
        cart[c] = code.encode("ascii")


_LANGLE, _RANGLE, _SLASH, _COLON = map(
    pp.Suppress, '<>/:'
)
_COMMENT = pp.Literal('--').suppress()

_tag_first = pp.Word(pp.alphas)
_tag_second = pp.match_previous_literal(_tag_first).suppress()
_new_line = pp.rest_of_line().copy().suppress()
_tag_start = pp.LineStart() + _COMMENT + _LANGLE + _tag_first + _RANGLE
_tag_end = pp.LineStart() + _COMMENT + _LANGLE + _SLASH + _tag_second + _RANGLE
_tag_line = pp.Group(_COMMENT + pp.Word(pp.nums) + _COLON + pp.Word(pp.hexnums))
_tag = _tag_start + pp.Group(pp.ZeroOrMore(_tag_line)) + _tag_end

_TEXTCHUNKS = {  # maxnumber, size per row, flip nibbles
    "TILES": (256, 32, True, ChunkID.TILES),
    "SPRITES": (256, 32, True, ChunkID.SPRITES),
    "MAP": (136, 240, True, ChunkID.MAP),
    "WAVES": (16, 16, True, ChunkID.WAVEFORM),
    "SFX": (64, 66, True, ChunkID.SAMPLES),
    "PATTERNS": (60, 64 * 3, True, ChunkID.PATTERNS),
    "TRACKS": (8, 51, True, ChunkID.MUSIC),
    "FLAGS": (2, 256, True, ChunkID.FLAGS),
    "SCREEN": (136, 120, True, ChunkID.SCREEN),
    "PALETTE": (2, 48, False, ChunkID.PALETTE),
}



from cmath import inf
from functools import singledispatch
import inspect
import itertools
import math
import pickle
import random
from typing import Any, Callable, Optional, Union, get_args, get_origin, get_type_hints
import tqdm
from pakettic import ast, parser
from dataclasses import replace


def _toBase26(num):
    s = ""
    while (num > 0):
        s += (chr(ord('z') - (num - 1) % 26))
        num -= 1
        num //= 26
    return s


_MINIFIED_NAMES = (name for i in itertools.count(start=1, step=1) if (name := _toBase26(i)) not in _RESERVED and name not in parser.keywords)
_FLIPPABLE_OPS = [">", "<", ">=", "<=", "~=", "=="]
_FLIPPED_OPS = ["<", ">", "<=", ">=", "~=", "=="]
_REORDERABLE_OPS = ["+", "-", "*", "/", "&", "|", "~"]
_REORDERABLE_RIGHT = [["+", "-"], ["+", "-"], ["*", "/"], ["*", "/"], ["&"], ["|"], ["~"]]
_REORDERABLE_LEFT = [True, False, True, False, True, True, True]
_EVALUABLE_OPS = {
    "+": lambda l, r: l + r,
    "-": lambda l, r: l - r,
    "*": lambda l, r: l * r,
    "/": lambda l, r: l / r,
    "//": lambda l, r: l // r,
    "%": lambda l, r: l % r,
    "^": lambda l, r: l ** r,
    "&": lambda l, r: int(l) & int(r) if int(l) == l and int(r) == r else None,
    "|": lambda l, r: int(l) | int(r) if int(l) == l and int(r) == r else None,
    "~": lambda l, r: int(l) ^ int(r) if int(l) == l and int(r) == r else None,
}

_LOWERS = list(chr(i) for i in range(ord('a'), ord('z') + 1))
_RESERVED = {'_G', 'BDR', 'SCN', 'BOOT', 'TIC', 'OVR', 'MENU', '_VERSION',
             'assert', 'btn', 'btnp', 'circ', 'circb', 'clip', 'cls',
             'collectgarbage', 'coroutine', 'debug', 'dofile', 'elli', 'ellib',
             'error', 'exit', 'fget', 'font', 'fset', 'getmetatable', 'ipairs',
             'key', 'keyp', 'line', 'load', 'loadfile', 'map', 'math', 'memcpy',
             'memset', 'mget', 'mouse', 'mset', 'music', 'next', 'package',
             'pairs', 'pcall', 'peek', 'peek1', 'peek2', 'peek4', 'pix', 'pmem',
             'poke', 'poke1', 'poke2', 'poke4', 'print', 'rawequal', 'rawget',
             'rawlen', 'rawset', 'rect', 'rectb', 'require', 'reset', 'select',
             'self', 'setmetatable', 'sfx', 'spr', 'str', 'string', 'sync', 'table',
             'textri', 'time', 'tonumber', 'tostring', 'trace', 'tri', 'trib',
             'tstamp', 'ttri', 'type', 'vbank', 'xpcall'}
_TWO = ast.Numeral(2)


def loads_to_funcs(root: ast.Node) -> ast.Node:
    """
    Transforms an abstract syntax tree by converting all load'...' into function()...end
        Parameters:
            root (ast.Node): Root of the abstract syntax tree
        Returns:
            new_root (ast.Node): Root of the new, transformed abstract syntax tree
    """
    def _trans(node: ast.Node) -> ast.Node:
        if isinstance(node, ast.Call) and node.func == ast.Name("load") and len(node.args) == 1 and isinstance(node.args[0], ast.LiteralString):
            body = parser.parse_string(node.args[0].value)
            return ast.Func(body=body, args=[])
        else:
            return node
    return apply_trans(root, _trans)


def minify(root: ast.Node) -> ast.Node:
    """
    Perform initial minification of the variable names
        Parameters:
            root (ast.Node): Root of the abstract syntax tree            
        Returns:
            new_root (ast.Node): Root of the new, mutated abstract syntax tree
    """
    new_root = pickle.loads(pickle.dumps(root))  # pickling/unpickling is faster than deepcopy
    changed_names = dict()
    changed_labels = dict()
    name_iter = iter(_MINIFIED_NAMES)
    label_iter = iter(_MINIFIED_NAMES)

    def _tryget(d: dict, key: str, it):
        if key in _RESERVED:
            return key
        new = d.get(key, None)
        if new is None:
            new = next(it)
        d[key] = new
        return new

    def _minify(node: ast.Node, parent: ast.Node, attr: str):
        if type(node) == ast.Name:
            node.id = _tryget(changed_names, node.id, name_iter)
        elif type(node) == ast.Label:
            node.name = _tryget(changed_labels, node.name, label_iter)
        elif type(node) == ast.Goto:
            node.target = _tryget(changed_labels, node.target, label_iter)

    visit(new_root, _minify)
    return new_root


def mutate(root: ast.Node, rand: random.Random) -> ast.Node:
    """
    Mutates an abstract syntax tree by making small modification to it,
    e.g. flipping binary operators or changing variable names
        Parameters:
            root (ast.Node): Root of the abstract syntax tree
            rand (random.Random): Random number generator to use
        Returns:
            new_root (ast.Node): Root of the new, mutated abstract syntax tree
    """
    new_root = pickle.loads(pickle.dumps(root))  # pickling/unpickling is faster than deepcopy
    mutations = []

    used_names = set()
    used_labels = set()

    def _check_mutations(node: ast.Node, parent: ast.Node, attr: str):
        if type(node) == ast.BinOp:
            try:
                i = _FLIPPABLE_OPS.index(node.op)

                def _mutation():
                    node.left, node.right = node.right, node.left
                    node.op = _FLIPPED_OPS[i]
                mutations.append(_mutation)
            except:
                pass
            try:
                i = _REORDERABLE_OPS.index(node.op)
                node2 = node
                while type(node2.left) == ast.BinOp and node2.left.op in _REORDERABLE_RIGHT[i]:
                    def _mutation(a=node2.left, b=node):
                        a.right, b.right = b.right, a.right
                        a.op, b.op = b.op, a.op
                    mutations.append(_mutation)
                    node2 = node2.left
                if _REORDERABLE_LEFT[i]:
                    def _mutation():
                        node2.left, node.right = node.right, node2.left
                    mutations.append(_mutation)
            except:
                pass
            if node.op == "*" and node.left == node.right:
                def _mutation():
                    node.op = "^"
                    node.right = _TWO
                mutations.append(_mutation)
            if node.op == "^" and node.right == _TWO:
                def _mutation():
                    node.op = "*"
                    node.right = pickle.loads(pickle.dumps(node.left))
                mutations.append(_mutation)
            if node.op in _EVALUABLE_OPS and type(node.left) == ast.Numeral and type(node.right) == ast.Numeral:
                value = _EVALUABLE_OPS[node.op](node.left.value, node.right.value)
                if value != None and int(value) == value:
                    def _mutation():
                        replace_node(parent, attr, ast.Numeral(int(value)), node)
                    mutations.append(_mutation)
        elif type(node) == ast.Name:
            used_names.add(node.id)
        elif type(node) == ast.Label:
            used_labels.add(node.name)
        elif type(node) == ast.Goto:
            used_labels.add(node.target)
        elif type(node) == ast.Alt:
            for i, _ in enumerate(node.alts[1:]):
                def _mutation(i=i):
                    node.alts[0], node.alts[i + 1] = node.alts[i + 1], node.alts[0]
                mutations.append(_mutation)
        elif type(node) == ast.Perm:
            if node.allow_reorder:
                for i in range(len(node.stats)):
                    for j in range(i + 1, len(node.stats)):
                        def _mutation(i=i, j=j):
                            node.stats[i], node.stats[j] = node.stats[j], node.stats[i]
                        mutations.append(_mutation)
        elif type(node) == ast.Func:
            if len(node.args) == 0 or (len(node.args) == 1 and type(node.args[0]) == ast.Ellipsis):
                def _mutation():
                    node.oneline = not node.oneline
                mutations.append(_mutation)
        elif type(node) == ast.ForRange:
            if node.step is None or node.step == ast.Numeral(1):
                def _mutation():
                    node.step = ast.Numeral(1) if node.step is None else None
                mutations.append(_mutation)
        elif type(node) == ast.Hint:
            def _mutation():
                node.double_quotes = not node.double_quotes
            mutations.append(_mutation)

            def _mutation2():
                node.no_hex = not node.no_hex
            mutations.append(_mutation2)

        if node.original != None and parent != None:
            def _mutation():
                replace_node(parent, attr, node.original, None)

    visit(new_root, _check_mutations)
    used_names = sorted(used_names.difference(_RESERVED))

    def var_repl(id_a, id_b):
        def _mut():
            def _repl(node, parent, attr):
                if type(node) == ast.Name:
                    if node.id == id_a:
                        node.id = id_b
                    elif node.id == id_b:
                        node.id = id_a
            visit(new_root, _repl)
        return _mut
    # if both a and b are in used_names, we have both copies of a,b and b,a in the list
    # but that's probably ok: swapping variables head to head is usually better idea
    # that swapping variables with new ones
    mutations.extend((var_repl(a, b) for a in used_names for b in _LOWERS if a != b))

    def label_repl(name_a, name_b):
        def _mut():
            def _repl(node, parent, attr):
                if type(node) == ast.Label:
                    if node.name == name_a:
                        node.name = name_b
                    elif node.name == name_b:
                        node.name = name_a
                if type(node) == ast.Goto:
                    if node.target == name_a:
                        node.target = name_b
                    elif node.target == name_b:
                        node.target = name_a
            visit(new_root, _repl)
        return _mut
    used_labels = sorted(used_labels)
    mutations.extend((label_repl(a, b) for a in used_labels for b in _LOWERS if a != b))
    if len(mutations) > 0:
        mutation = rand.choice(mutations)
        mutation()
    return new_root


def replace_node(parent: ast.Node, attr: str, node: ast.Node, old: ast.Node):
    """
    Replaces one node in the abstract syntax tree with another.
    Adds a reference to the old node from the new node's 'original' attribute.
        Parameters:
            parent (ast.Node): Root of the abstract syntax tree
            attr (str): parent attribute which references the node being replaced. Can be "attr.#" to replace an element of a list.
            node (ast.Node): new node
            old (ast.Node): old node that is being replaced
    """
    node.original = old
    if "." in attr:
        a, i = attr.split(".")
        getattr(parent, a)[int(i)] = node
    else:
        setattr(parent, attr, node)


def anneal(state: Any, cost_func: Callable[[Any, int], int], steps: int, start_temp: float, end_temp: float, seed: int, mutate_func: Callable[[Any], Any] = mutate) -> Any:
    """
    Perform simulated annealing optimization, using exponential temperature schedule.
    See https://en.wikipedia.org/wiki/Simulated_annealing
        Parameters:
            state: Starting state for the optimization
            cost_func (Callable[[Any, int], int]): Callback function, with the first parameter the state and second parameter the cost of best solution so far
            steps (int): how many steps the optimization algorithms takes
            start_temp (float): starting temperature for the optimization
            end_temp (float): end temperature for the optimization
            mutate_func (Callable[[Any], Any]): Callback function state -> state that performs a small mutation in the code
        Returns:
            best (Any): The best solution found
    """
    current_cost = cost_func(state, inf)
    best_cost = current_cost
    best = state
    r = random.Random(seed)  # deterministic seed, to have deterministic results
    bar = tqdm.tqdm(range(steps), position=1, leave=False)
    for i in bar:
        alpha = i / (steps - 1)
        temp = math.exp((1 - alpha) * math.log(start_temp) + alpha * math.log(end_temp))
        candidate = mutate_func(state, r)
        cand_cost = cost_func(candidate, best_cost)
        if cand_cost < current_cost or math.exp(-(cand_cost - current_cost) / temp) >= r.random():
            current_cost = cand_cost
            state = candidate
        if cand_cost < best_cost:
            best_cost = cand_cost
            best = candidate
        bar.set_description(f"B:{best_cost} C:{current_cost} A:{cand_cost} T: {temp:.1f}")
        if best_cost <= 0:
            break
    return best


def lahc(state: Any, cost_func: Callable[[Any, int], int], steps: int, list_length: int, init_margin: int, seed: int, mutate_func: Callable[[Any], Any] = mutate) -> Any:
    """
    Optimize a function using Late Acceptance Hill Climbing
    See https://arxiv.org/pdf/1806.09328.pdf
        Parameters:
            state: Starting state for the optimization
            cost_func (Callable[[Any, int], int]): Callback function, with the first parameter the state and second parameter the cost of best solution so far
            steps (int): how many steps the optimization algorithms takes
            list_length (int): length of the history in the algorithm
            init_margin (int): how much margin, in bytes, to add to the initial best cost
            mutate_func (Callable[[Any], Any]): Callback function state -> state that performs a small mutation in the code
        Returns:
            best (Any): The best solution found
    """
    current_cost = cost_func(state, inf)
    best_cost = current_cost
    history = [best_cost + init_margin] * list_length
    best = state
    r = random.Random(seed)  # deterministic seed, to have deterministic results
    bar = tqdm.tqdm(range(steps), position=1, leave=False)
    for i in bar:
        candidate = mutate_func(state, r)
        cand_cost = cost_func(candidate, best_cost)
        v = i % list_length
        if cand_cost < history[v] or cand_cost <= current_cost:
            current_cost = cand_cost
            state = candidate
        if current_cost < history[v]:
            history[v] = current_cost
        if cand_cost < best_cost:
            best_cost = cand_cost
            best = candidate
        bar.set_description(f"B:{best_cost} C:{current_cost} A:{cand_cost}")
        if best_cost <= 0:
            break
    return best


def dlas(state: Any, cost_func: Callable[[Any, int], int], steps: int, list_length: int, init_margin: int, seed: int, mutate_func: Callable[[Any], Any] = mutate) -> Any:
    """
    Optimize a function using Diversified Late Acceptance Search
    See https://arxiv.org/pdf/1806.09328.pdf
        Parameters:
            state: Starting state for the optimization
            cost_func (Callable[[Any, int], int]): Callback function, with the first parameter the state and second parameter the cost of best solution so far
            steps (int): how many steps the optimization algorithms takes
            list_length (int): length of the history in the algorithm
            init_margin (int): how much margin, in bytes, to add to the initial best cost
            mutate_func (Callable[[Any], Any]): Callback function state -> state that performs a small mutation in the code
        Returns:
            best (Any): The best solution found
    """
    current_cost = cost_func(state, inf)
    best_cost = current_cost
    cost_max = best_cost + init_margin
    history = [cost_max] * list_length
    N = list_length
    best = state
    r = random.Random(seed)  # deterministic seed, to have deterministic results
    bar = tqdm.tqdm(range(steps), position=1, leave=False)
    for i in bar:
        prev_cost = current_cost
        candidate = mutate_func(state, r)
        cand_cost = cost_func(candidate, best_cost)
        v = i % list_length
        if cand_cost == current_cost or cand_cost < cost_max:
            current_cost = cand_cost
            state = candidate
        if current_cost > history[v]:
            history[v] = current_cost
        elif current_cost < history[v] and current_cost < prev_cost:
            if history[v] == cost_max:
                N -= 1
            history[v] = current_cost
            if N <= 0:
                cost_max = max(history)
                N = history.count(cost_max)
        if cand_cost < best_cost:
            best_cost = cand_cost
            best = candidate
        bar.set_description(f"B:{best_cost} C:{current_cost} M:{cost_max} A:{cand_cost}")
        if best_cost <= 0:
            break
    return best


def apply_trans(node: ast.Node, trans: Callable[[ast.Node], ast.Node]) -> ast.Node:
    """
    Applies a transformation to an abstract syntax tree recursively
        Parameters:
            node (ast.Node): Root of the syntax tree to transform
            trans (Callable[[ast.Node], ast.Node]): Callback function that takes a node and returns a transformed node
        Returns:
            new_node (ast.Node): Root of the new, transformed abstract syntax tree
    """
    if node is None:
        return trans(node)
    replaces = dict()
    try:
        for name, typehint in get_type_hints(node).items():
            attr = getattr(node, name)
            origin = get_origin(typehint)
            args = get_args(typehint)
            if origin is Optional:
                if attr is None:
                    continue
                origin = args
            if origin is list and len(args) > 0 and issubclass(args[0], ast.Node):
                replaces[name] = [apply_trans(e, trans) for e in attr]
            elif inspect.isclass(typehint) and issubclass(typehint, ast.Node):
                replaces[name] = apply_trans(attr, trans)
    except TypeError:
        pass
    return trans(replace(node, **replaces))


@singledispatch
def visit(node: ast.Node, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    """
    Visit each node of an abstract syntax tree recursively
        Parameters:
            node (ast.Node): A (root) node of the syntax tree to visit
            visitor (Callable[[ast.Node, ast.Node, str], None]): Callback function called for each node
    """
    if node != None:
        visitor(node, parent, attr)


@ visit.register
def _(node: ast.Block, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    for s in node.stats:
        visit(s, visitor)


@ visit.register
def _(node: ast.Return, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    for i, e in enumerate(node.exps):
        visit(e, visitor, node, f"exps.{i}")


@ visit.register
def _(node: ast.Do, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    visit(node.block, visitor, node, "block")


@ visit.register
def _(node: ast.Assign, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    for i, t in enumerate(node.targets):
        visit(t, visitor, node, f"targets.{i}")
    for i, v in enumerate(node.values):
        visit(v, visitor, node, f"values.{i}")


@ visit.register
def _(node: ast.While, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    visit(node.condition, visitor, node, "condition")
    visit(node.block, visitor, node, "block")


@ visit.register
def _(node: ast.Repeat, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    visit(node.block, visitor, node, "block")
    visit(node.condition, visitor, node, "condition")


@ visit.register
def _(node: ast.ForRange, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    visit(node.var, visitor, node, "var")
    visit(node.lb, visitor, node, "lb")
    visit(node.ub, visitor, node, "ub")
    visit(node.step, visitor, node, "step")
    visit(node.body, visitor, node, "body")


@ visit.register
def _(node: ast.ForIn, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    for i, n in enumerate(node.names):
        visit(n, visitor, node, f"names.{i}")
    for i, e in enumerate(node.exps):
        visit(e, visitor, node, f"exps.{i}")
    visit(node.body, visitor, node, "body")


@ visit.register
def _(node: ast.Local, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    for i, t in enumerate(node.targets):
        visit(t, visitor, node, f"targets.{i}")
    if node.values is not None:
        for i, v in enumerate(node.values):
            visit(v, visitor, node, f"values.{i}")


@ visit.register
def _(node: ast.Func, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    for i, a in enumerate(node.args):
        visit(a, visitor, node, f"args.{i}")
    visit(node.body, visitor, node, "body")


@ visit.register
def _(node: ast.Index, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    visit(node.obj, visitor, node, "obj")
    visit(node.item, visitor, node, "item")


@ visit.register
def _(node: ast.Table, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    for i, f in enumerate(node.fields):
        visit(f, visitor, node, f"fields.{i}")


@ visit.register
def _(node: ast.Field, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    visit(node.key, visitor, node, "key")
    visit(node.value, visitor, node, "value")


@ visit.register(ast.MethodCall)
def _(node: ast.MethodCall, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    visit(node.value, visitor, node, "value")
    for i, a in enumerate(node.args):
        visit(a, visitor, node, f"args.{i}")


@ visit.register(ast.Call)
def _(node: ast.Call, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    visit(node.func, visitor, node, "func")
    for i, a in enumerate(node.args):
        visit(a, visitor, node, f"args.{i}")


@ visit.register
def _(node: ast.If, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    visit(node.test, visitor, node, "test")
    visit(node.body, visitor, node, "body")
    visit(node.orelse, visitor, node, "orelse")


@ visit.register
def _(node: ast.BinOp, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    visit(node.left, visitor, node, "left")
    visit(node.right, visitor, node, "right")


@ visit.register
def _(node: ast.UnaryOp, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    visit(node.operand, visitor, node, "operand")


@ visit.register
def _(node: ast.Alt, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    for i, a in enumerate(node.alts):
        visit(a, visitor, node, f"alts.{i}")


@ visit.register
def _(node: ast.Perm, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    for i, s in enumerate(node.stats):
        visit(s, visitor, node, f"stats.{i}")


@ visit.register
def _(node: ast.Hint, visitor: Callable[[ast.Node, ast.Node, str], None], parent: ast.Node = None, attr: str = None):
    visitor(node, parent, attr)
    visit(node.block, visitor, node, "block")





from dataclasses import dataclass
from functools import singledispatch
import typing
import re
from pakettic import ast


def format(node: ast.Node, pretty: bool = False, no_load: bool = False) -> str:
    return Formatter(pretty=pretty, no_load=no_load).format(node)


_hexy = re.compile(r'[0-9a-fxA-FXpP\.]').search
_alphaunder = re.compile('[_a-zA-Z]').search
_alphanumunder = re.compile('[_a-zA-Z0-9]').search
_single_quote_translation = str.maketrans({"\n": r"\n",
                                           "\t": r"\t",
                                           "\f": r"\f",
                                           "\r": r"\r",
                                           "\\": r"\\",
                                           "'": r"\'"})
_double_quote_translation = str.maketrans({"\n": r"\n",
                                           "\t": r"\t",
                                           "\f": r"\f",
                                           "\r": r"\r",
                                           "\\": r"\\",
                                           "\"": r'\"'})


@dataclass
class Formatter:
    indent: int = 0
    double_quotes: bool = False
    pretty: bool = False
    no_hex: bool = False
    no_load: bool = False

    def format(self, node: ast.Node) -> str:
        tokens = _traverse(node, self)
        return ''.join(self.__addspaces(tokens))

    @property
    def __quote(self):
        return '"' if self.double_quotes else "'"

    def escape(self, s: str):
        return f"{self.__quote}{s.translate(_double_quote_translation if self.double_quotes else _single_quote_translation)}{self.__quote}"

    def __addspaces(self, tokens: typing.Iterable[str]):
        prevtoken = ' '
        for token in tokens:
            if type(token) is str:
                if len(token) == 0:
                    continue
                if type(prevtoken) is str and bool(_alphaunder(prevtoken[0])) and bool(_alphanumunder(token[0])):
                    # the previous token was word, and the next continues with a character that might be confused with it
                    yield ' '
                elif type(prevtoken) is ast.Numeral and bool(_hexy(token[0])):
                    yield ' '  # the previous token was numeral and the next starts with something that be confused with a hex or a decimal point
                yield token
            elif type(token) is ast.Numeral:
                strnumeral = str(token)
                if strnumeral[0] != '.' and type(prevtoken) is str and bool(_alphaunder(prevtoken[0])):
                    # the previous token was word, and the next continues with a character that might be confused with it
                    yield ' '
                yield strnumeral
            prevtoken = token


# This used to be a singledispatchmethod of Formatter, but for some reason, @singledispatchmethod are far slower than
# @singledispatch
@ singledispatch
def _traverse(node: ast.Node, fmt: Formatter):
    # raise TypeError("_print encounted unknown ast.node")
    yield str(node)


@ _traverse.register
def _(node: ast.Block, fmt: Formatter):
    for n in node.stats:
        if fmt.pretty:
            yield '  ' * fmt.indent
        yield from _traverse(n, fmt)
        if fmt.pretty:
            yield '\n'


@ _traverse.register
def _(node: ast.Return, fmt: Formatter):
    yield 'return'
    for i, v in enumerate(node.exps):
        if i > 0:
            yield ','
        yield from _traverse(v, fmt)


@ _traverse.register
def _(node: ast.Perm, fmt: Formatter):
    if fmt.pretty:
        if node.allow_reorder:
            yield '--{'
        else:
            yield '--{!'
        yield '\n'
        yield '  ' * fmt.indent
    for i, n in enumerate(node.stats):
        if fmt.pretty and i > 0:
            yield '  ' * fmt.indent
        yield from _traverse(n, fmt)
        if fmt.pretty and i < len(node.stats) - 1:
            yield '\n'
    if fmt.pretty:
        yield '\n'
        yield '  ' * fmt.indent
        yield '--}'


@ _traverse.register
def _(node: ast.Do, fmt: Formatter):
    yield 'do'
    if fmt.pretty:
        yield '\n'
    fmt.indent += 1
    yield from _traverse(node.block, fmt)
    fmt.indent -= 1
    if fmt.pretty:
        yield '  ' * fmt.indent
    yield 'end'


@ _traverse.register
def _(node: ast.Assign, fmt: Formatter):
    for i, v in enumerate(node.targets):
        if i > 0:
            yield ','
        yield from _traverse(v, fmt)
    yield '='
    for i, v in enumerate(node.values):
        if i > 0:
            yield ','
        yield from _traverse(v, fmt)


@ _traverse.register
def _(node: ast.Label, fmt: Formatter):
    yield '::'
    yield node.name
    yield '::'


@ _traverse.register
def _(node: ast.Break, fmt: Formatter):
    yield 'break'


@ _traverse.register
def _(node: ast.LiteralString, fmt: Formatter):
    yield fmt.escape(node.value)


@ _traverse.register
def _(node: ast.Goto, fmt: Formatter):
    yield 'goto'
    yield node.target


@ _traverse.register
def _(node: ast.Boolean, fmt: Formatter):
    yield 'true' if node.value else 'false'


@ _traverse.register
def _(node: ast.Ellipsis, fmt: Formatter):
    yield '...'


@ _traverse.register
def _(node: ast.Nil, fmt: Formatter):
    yield 'nil'


@ _traverse.register
def _(node: ast.While, fmt: Formatter):
    yield 'while'
    yield from _traverse(node.condition, fmt)
    yield 'do'
    if fmt.pretty:
        yield '\n'
    fmt.indent += 1
    yield from _traverse(node.block, fmt)
    fmt.indent -= 1
    if fmt.pretty:
        yield '  ' * fmt.indent
    yield 'end'


@ _traverse.register
def _(node: ast.Repeat, fmt: Formatter):
    yield 'repeat'
    if fmt.pretty:
        yield '\n'
    fmt.indent += 1
    yield from _traverse(node.block, fmt)
    fmt.indent -= 1
    yield 'until'
    yield from _traverse(node.condition, fmt)


@ _traverse.register
def _(node: ast.ForRange, fmt: Formatter):
    yield 'for'
    yield from _traverse(node.var, fmt)
    yield '='
    yield from _traverse(node.lb, fmt)
    yield ','
    yield from _traverse(node.ub, fmt)
    if node.step is not None:
        yield ','
        yield from _traverse(node.step, fmt)
    yield 'do'
    if fmt.pretty:
        yield '\n'
    fmt.indent += 1
    yield from _traverse(node.body, fmt)
    fmt.indent -= 1
    if fmt.pretty:
        yield '  ' * fmt.indent
    yield 'end'


@ _traverse.register
def _(node: ast.ForIn, fmt: Formatter):
    yield 'for'
    for i, v in enumerate(node.names):
        if i > 0:
            yield ','
        yield from _traverse(v, fmt)
    yield 'in'
    for i, v in enumerate(node.exps):
        if i > 0:
            yield ','
        yield from _traverse(v, fmt)
    yield 'do'
    if fmt.pretty:
        yield '\n'
    fmt.indent += 1
    yield from _traverse(node.body, fmt)
    fmt.indent -= 1
    if fmt.pretty:
        yield '  ' * fmt.indent
    yield 'end'


@ _traverse.register
def _(node: ast.Local, fmt: Formatter):
    yield 'local'
    for i, v in enumerate(node.targets):
        if i > 0:
            yield ','
        yield from _traverse(v, fmt)
    if node.values is not None:
        yield '='
        for i, v in enumerate(node.values):
            if i > 0:
                yield ','
            yield from _traverse(v, fmt)


@ _traverse.register
def _(node: ast.Func, fmt: Formatter):
    if (len(node.args) == 0 or (len(node.args) == 1 and type(node.args[0]) == ast.Ellipsis)) and not fmt.pretty and not fmt.no_load and node.oneline:
        fmt.indent += 1
        fmt2 = Formatter(fmt.indent + 1, double_quotes=not fmt.double_quotes, pretty=False, no_load=fmt.no_load)
        s = fmt2.format(node.body)
        yield 'load'
        yield fmt.escape(s)
        fmt.indent -= 1
    else:
        yield 'function'
        yield '('
        for i, v in enumerate(node.args):
            if i > 0:
                yield ','
            yield from _traverse(v, fmt)
        yield ')'
        if fmt.pretty:
            yield '\n'
        fmt.indent += 1
        yield from _traverse(node.body, fmt)
        fmt.indent -= 1
        if fmt.pretty:
            yield '  ' * fmt.indent
        yield 'end'


@ _traverse.register
def _(node: ast.Index, fmt: Formatter):
    require_parentheses = type(node.obj) is not ast.Name and \
        type(node.obj) is not ast.Index and type(node.obj) is not ast.MethodCall
    if require_parentheses:
        yield '('
    yield from _traverse(node.obj, fmt)
    if require_parentheses:
        yield ')'
    if type(node.item) is ast.LiteralString:
        yield '.'
        yield node.item.value
    else:
        yield '['
        yield from _traverse(node.item, fmt)
        yield ']'


@ _traverse.register
def _(node: ast.Table, fmt: Formatter):
    yield '{'
    for i, f in enumerate(node.fields):
        if i > 0:
            yield ','
        yield from _traverse(f, fmt)
    yield '}'


@ _traverse.register
def _(node: ast.Field, fmt: Formatter):
    if node.key is not None:
        if type(node.key) is ast.LiteralString:
            yield node.key.value
            yield '='
        else:
            yield '['
            yield from _traverse(node.key, fmt)
            yield ']'
            yield '='
    yield from _traverse(node.value, fmt)


@ _traverse.register
def _(node: ast.Call, fmt: Formatter):
    yield from _traverse(node.func, fmt)
    if len(node.args) == 1 and (type(node.args[0]) is ast.Table or type(node.args[0]) is ast.LiteralString):
        yield from _traverse(node.args[0], fmt)
    else:
        yield '('
        for i, v in enumerate(node.args):
            if i > 0:
                yield ','
            yield from _traverse(v, fmt)
        yield ')'


@ _traverse.register
def _(node: ast.MethodCall, fmt: Formatter):
    yield from _traverse(node.value, fmt)
    yield ':'
    yield from _traverse(node.method, fmt)
    yield '('
    for i, v in enumerate(node.args):
        if i > 0:
            yield ','
        yield from _traverse(v, fmt)
    yield ')'


@ _traverse.register
def _(node: ast.If, fmt: Formatter):
    yield 'if'
    yield from _traverse(node.test, fmt)
    yield 'then'
    fmt.indent += 1
    if fmt.pretty:
        yield '\n'
    yield from _traverse(node.body, fmt)
    fmt.indent -= 1
    while node.orelse is not None and len(node.orelse.stats) == 1 and type(node.orelse.stats[0]) == ast.If:
        node = node.orelse.stats[0]
        if fmt.pretty:
            yield '  ' * fmt.indent
        yield 'elseif'
        yield from _traverse(node.test, fmt)
        yield 'then'
        fmt.indent += 1
        if fmt.pretty:
            yield '\n'
        yield from _traverse(node.body, fmt)
        fmt.indent -= 1
    if node.orelse is not None:
        if fmt.pretty:
            yield '  ' * fmt.indent
        yield 'else'
        if fmt.pretty:
            yield '\n'
        fmt.indent += 1
        yield from _traverse(node.orelse, fmt)
        fmt.indent -= 1
    if fmt.pretty:
        yield '  ' * fmt.indent
    yield 'end'


@ _traverse.register
def _(node: ast.Name, fmt: Formatter):
    yield node.id


@ _traverse.register
def _(node: ast.BinOp, fmt: Formatter):
    if node.op == '^':
        if node.left.precedence >= node.precedence:
            yield '('
        yield from _traverse(node.left, fmt)
        if node.left.precedence >= node.precedence:
            yield ')'
        yield node.op
        if node.right.precedence > node.precedence:
            yield '('
        yield from _traverse(node.right, fmt)
        if node.right.precedence > node.precedence:
            yield ')'
    else:
        if node.left.precedence > node.precedence:
            yield '('
        yield from _traverse(node.left, fmt)
        if node.left.precedence > node.precedence:
            yield ')'
        yield node.op
        if node.right.precedence >= node.precedence:
            yield '('
        yield from _traverse(node.right, fmt)
        if node.right.precedence >= node.precedence:
            yield ')'


@ _traverse.register
def _(node: ast.UnaryOp, fmt: Formatter):
    yield node.op
    if node.operand.precedence > node.precedence:
        yield '('
    yield from _traverse(node.operand, fmt)
    if node.operand.precedence > node.precedence:
        yield ')'


@ _traverse.register
def _(node: ast.Alt, fmt: Formatter):
    yield from _traverse(node.alts[0], fmt)
    if fmt.pretty and len(node.alts) > 0:
        for i, v in enumerate(node.alts[1:]):
            yield '--|'
            yield from _traverse(v, fmt)
            if i == len(node.alts) - 2:
                yield '\n'
                yield '  ' * fmt.indent


@ _traverse.register
def _(node: ast.Numeral, fmt: Formatter):
    # We treat numerals a bit differently, to see if they need spaces before, so
    # we yield them as Numerals, not strs
    if fmt.no_hex and node.hex and node.fractional == 0 and node.exponent == 0:
        yield ast.Numeral(node.whole)
    else:
        yield node


@ _traverse.register
def _(node: ast.Hint, fmt: Formatter):
    prev_double_quotes = fmt.double_quotes
    prev_no_hex = fmt.no_hex
    fmt.double_quotes = node.double_quotes
    fmt.no_hex = node.no_hex
    yield from _traverse(node.block, fmt)
    fmt.double_quotes = prev_double_quotes
    fmt.no_hex = prev_no_hex



from xml.dom.expatbuilder import DOCUMENT_NODE
from pakettic import ast
import pyparsing as pp
import functools

pp.ParserElement.enablePackrat()

LBRACK, RBRACK, LBRACE, RBRACE, LPAR, RPAR, EQ, COMMA, SEMI, COLON, PERIOD, VLINE = map(
    pp.Suppress, '[]{}()=,;:.|'
)
keywords = """\
    return break do end while if then elseif else for in function local repeat until nil false true and or not goto
    """.split()
keyword_pats = {
    k.upper(): pp.Regex(k + r"(?![a-zA-Z_])").suppress()
    for k in keywords
}
vars().update(keyword_pats)
any_keyword = pp.MatchFirst(keyword_pats.values()).setName("<keyword>")

# Name
Name = ~any_keyword + pp.Word(pp.alphas + "_", pp.alphanums + "_")

# LiteralString
short_literal_string = (pp.QuotedString(
    '"', escChar='\\') | pp.QuotedString('\'', escChar='\\')).setParseAction(lambda toks: ast.LiteralString(toks[0]))
long_literal_string = pp.Regex(r"\[(=*)\[(?P<str>[\s\S]*?)\]\1\]")
long_literal_string.setParseAction(lambda t: ast.LiteralString(
    t["str"][1:] if t["str"].startswith("\n") else t["str"]))
LiteralString = short_literal_string | long_literal_string

# Numeral
base10 = pp.Regex(r"(?P<whole>\d+)(?:\.(?P<frac>\d*))?(?:[eE](?P<exp>[+-]?\d+))?")
base10.set_parse_action(lambda t: ast.Numeral(int(t["whole"] or 0), int(
    t["frac"][::-1]) if t["frac"] else 0, int(t["exp"] or 0)))

frac10 = pp.Regex(r"\.(?P<frac>\d+)(?:[eE](?P<exp>[+-]?\d+))?")
frac10.set_parse_action(lambda t: ast.Numeral(0, int(t["frac"][::-1] or 0), int(t["exp"] or 0)))

base16 = pp.Regex(r"0[xX](?P<whole>[0-9a-fA-F]+)(?:\.(?P<frac>[0-9a-fA-F]*))?(?:[pP](?P<exp>[+-]?\d+))?")
base16.set_parse_action(lambda t: ast.Numeral(int(t["whole"], 16) if t["whole"]
                        else 0, int(t["frac"][::-1], 16) if t["frac"] else 0, int(t["exp"] or 0), hex=True))

frac16 = pp.Regex(r"0[xX]\.(?P<frac>[0-9a-fA-F]*)(?:[pP](?P<exp>[+-]?\d+))?")
frac16.set_parse_action(lambda t: ast.Numeral(0, int(t["frac"][::-1], 16) if t["frac"] else 0, int(t["exp"] or 0), hex=True))

# warning: order is important. base16 have to come before base10 as 0x would be matched as 0
Numeral = base16 | frac16 | base10 | frac10

# The actual grammar
stat = pp.Forward()
retstat = pp.Forward()
varlist = pp.Forward()
explist = pp.Forward()
functioncall = pp.Forward()
label = pp.Forward()
exp = pp.Forward()
namelist = pp.Forward()
funcname = pp.Forward()
funcbody = pp.Forward()
var = pp.Forward()
functiondef = pp.Forward()
prefixexp = pp.Forward()
tableconstructor = pp.Forward()
args = pp.Forward()
parlist = pp.Forward()
fieldlist = pp.Forward()
fieldsep = pp.Forward()
field = pp.Forward()

# block ::= {stat} [retstat]
block = pp.Group(stat[0, ...] + retstat[0, 1], aslist=True)
block.set_parse_action(lambda t: ast.Block(t[0]))
chunk = block

# stat ::=  ‘;’ |
#        varlist ‘=’ explist |
#        functioncall |
#        label |
#        break |
#        goto Name |
#        do block end |
#        while exp do block end |
#        repeat block until exp |
#        if exp then block {elseif exp then block} [else block] end |
#        for Name ‘=’ exp ‘,’ exp [‘,’ exp] do block end |
#        for namelist in explist do block end |
#        function funcname funcbody |
#        local function Name funcbody |
#        local namelist [‘=’ explist]
assign = varlist + EQ + explist
assign.set_parse_action(lambda toks: ast.Assign(toks[0], toks[1]))
break_ = BREAK.set_parse_action(lambda toks: ast.Break())
goto = (GOTO - Name).set_parse_action(lambda toks: ast.Goto(toks[0]))
doblock = DO - block - END
doblock.set_parse_action(lambda t: ast.Do(t[0]))
permblockstart = pp.Regex(r"--{[!]?")
permblockstart.set_parse_action(lambda t:
                                len(t[0]) < 4)
permblock = permblockstart + pp.Group(stat[0, ...], aslist=True) + pp.Literal('--}').suppress()
permblock.set_parse_action(lambda t: ast.Perm(t[1], allow_reorder=t[0]))
while_ = WHILE - exp - DO - block - END
while_.set_parse_action(lambda toks: ast.While(toks[0], toks[1]))
repeat = REPEAT - block - UNTIL - exp
repeat.set_parse_action(lambda toks: ast.Repeat(condition=toks[1], block=toks[0]))
for_range = FOR + Name + EQ - exp - COMMA - exp - (COMMA - exp)[0, 1] - DO - block - END
for_range.set_parse_action(lambda t: ast.ForRange(ast.Name(t[0]), t[1], t[2], None, t[3]) if len(
    t) < 5 else ast.ForRange(ast.Name(t[0]), t[1], t[2], t[3], t[4]))
if_ = IF - exp + THEN + block + \
    pp.Group(pp.ZeroOrMore(pp.Group(ELSEIF - exp + THEN + block))) + \
    pp.Optional(ELSE - block, default=None) + END
# The AST does not know anything about elseif; split them into else if
if_.set_parse_action(lambda toks: functools.reduce(lambda x, y: ast.If(x.test, body=x.body, orelse=ast.Block([ast.If(
    test=y[0], body=y[1], orelse=x.orelse)])), reversed(toks[2]), ast.If(test=toks[0], body=toks[1], orelse=toks[3])))
for_in = FOR + pp.Group(namelist, aslist=True) + IN - \
    explist - DO - block - END
for_in.set_parse_action(lambda toks: ast.ForIn(toks[0], toks[1], toks[2]))
local_var = LOCAL + pp.Group(namelist, aslist=True) - (EQ + explist)[0, 1]
local_var.set_parse_action(lambda toks: ast.Local(toks[0], None) if len(toks) < 2 else ast.Local(toks[0], toks[1]))
func_vanilla_def = FUNCTION + funcname + funcbody
func_vanilla_def.set_parse_action(lambda toks: ast.Assign([toks[0]], [toks[1]]))
# function t.a.b.c:f (params) body end is syntactic sugar for t.a.b.c.f = function (self, params) body end
func_member_def = FUNCTION + funcname + COLON - Name - funcbody
func_member_def.set_parse_action(lambda t: ast.Assign(
    [ast.Index(t[0], ast.LiteralString(t[1]))],
    [ast.Func(args=[ast.Name('self')] + t[2].args, body=t[2].body)]
))
func_def = func_vanilla_def | func_member_def
local_func_def = LOCAL + FUNCTION - Name - funcbody
local_func_def.set_parse_action(lambda toks: ast.Local([ast.Name(toks[0])], [toks[1]]))

stat <<= SEMI | \
    label | \
    break_ | \
    goto | \
    doblock | \
    permblock | \
    while_ | \
    repeat | \
    for_range | \
    if_ | \
    for_in | \
    local_var | \
    func_def | \
    assign | \
    local_func_def | \
    functioncall

# retstat ::= return [explist] [‘;’]
retstat <<= RETURN - pp.Optional(explist, default=[]) - SEMI[0, 1]
retstat.set_parse_action(lambda toks: ast.Return(exps=toks[0]))

# label ::= ‘::’ Name ‘::’
label <<= pp.Literal('::').suppress() - Name - pp.Literal('::').suppress()
label.set_parse_action(lambda toks: ast.Label(toks[0]))

# funcname ::= Name {‘.’ Name} [‘:’ Name]
funcname <<= Name - (PERIOD - Name)[0, ...]
funcname.set_parse_action(lambda t: functools.reduce(lambda x, y: ast.Index(x, ast.LiteralString(y)), t[1:], ast.Name(t[0])))
# the [‘:’ Name] case is handled separately

# varlist ::= var {‘,’ var}
varlist <<= pp.Group(var - (COMMA - var)[0, ...], aslist=True)

# namelist ::= Name {‘,’ Name}
_Name = Name.copy().set_parse_action(lambda t: ast.Name(t[0]))
namelist <<= _Name - (COMMA - _Name)[0, ...]

# explist ::= exp {‘,’ exp}
explist <<= pp.Group(exp - (COMMA - exp)[0, ...], aslist=True)

# exp ::=  nil | false | true | Numeral | LiteralString | ‘...’ | functiondef | prefixexp | tableconstructor | exp binop exp | unop exp


left_assoc = lambda t: functools.reduce(lambda x, y: ast.BinOp(x, y[0], y[1]), zip(t[0][1::2], t[0][2::2]), t[0][0])
right_assoc = lambda t: functools.reduce(lambda x, y: ast.BinOp(y[1], y[0], x), zip(t[0][-2::-2], t[0][-3::-2]), t[0][-1])
alt = lambda t: ast.Alt(t.asList()) if len(t) > 1 else t[0]


def unaryAction(t):
    return functools.reduce(lambda x, y: ast.UnaryOp(op=y, operand=x), t[-2::-1], t[-1])


nil = pp.Keyword("nil").set_parse_action(lambda: ast.Nil())
false = pp.Keyword("false").set_parse_action(lambda: ast.Boolean(False))
true = pp.Keyword("true").set_parse_action(lambda: ast.Boolean(True))
ellipsis = pp.Literal("...").set_parse_action(lambda: ast.Ellipsis())
# The power operator is weird: it binds more strongly than unary to the
# thing on its left, but less strongly than unary to its right. So, we
# need to treat it separately
expliteral = nil | false | true | Numeral | LiteralString | ellipsis | functiondef | prefixexp | tableconstructor | LPAR + exp + RPAR
altexp = expliteral + (pp.Literal('--|').suppress() - expliteral)[0, ...]
altexp.set_parse_action(alt)
powerexp = pp.Forward()
unaryexp = pp.oneOf('not # - ~')[0, ...] + powerexp
unaryexp.set_parse_action(unaryAction)
powerexp <<= altexp + (pp.Literal('^').suppress() - unaryexp)[0, 1]
powerexp.set_parse_action(lambda t: ast.BinOp(left=t[0], op="^", right=t[1]) if len(t) > 1 else t[0])
exp <<= pp.infixNotation(
    unaryexp,
    [
        (pp.oneOf('* / // %'), 2, pp.OpAssoc.LEFT, left_assoc),
        (pp.oneOf('+ -'), 2, pp.OpAssoc.LEFT, left_assoc),
        ('..', 2, pp.OpAssoc.LEFT, left_assoc),
        (pp.oneOf('<< >>'), 2, pp.OpAssoc.LEFT, left_assoc),
        ('&', 2, pp.OpAssoc.LEFT, left_assoc),
        ('~', 2, pp.OpAssoc.LEFT, left_assoc),
        ('|', 2, pp.OpAssoc.LEFT, left_assoc),
        (pp.oneOf('< > <= >= ~= ==', left_assoc),
         2, pp.OpAssoc.LEFT, left_assoc),
        (pp.Regex(r"and(?![a-zA-Z_])"), 2, pp.OpAssoc.LEFT, left_assoc),  # can't use AND as it suppresses the token
        (pp.Regex(r"or(?![a-zA-Z_])"), 2, pp.OpAssoc.LEFT, left_assoc),  # can't use OR as it suppresses the token
    ]
)

# args: := ‘(’ [explist] ‘)’ | tableconstructor | LiteralString
args <<= LPAR + pp.Optional(explist, default=[]) + RPAR | (tableconstructor | LiteralString).set_parse_action(lambda t: [[t[0]]])


# prefixexp: := var | functioncall | ‘(’ exp ‘)’
# prefixexp has a nasty left recursive grammar. we need to implement the grammar slightly differently to avoid it
postfix = args.copy().set_parse_action(lambda t: ast.Call(None, t[0])) | \
    (COLON + Name + args).set_parse_action(lambda t: ast.MethodCall(None, t[0], t[1])) | \
    (LBRACK + exp + RBRACK).set_parse_action(lambda t: ast.Index(None, t[0])) | \
    (PERIOD + Name).set_parse_action(lambda t: ast.Index(None, ast.LiteralString(t[0])))
prefixexp <<= (
    Name.copy().set_parse_action(lambda toks: ast.Name(id=toks[0])) |
    LPAR + exp + RPAR) + pp.ZeroOrMore(postfix)
prefixexp.set_parse_action(lambda t: functools.reduce(
    lambda x, y: ast.Call(x, y.args) if type(y) is ast.Call else
    ast.MethodCall(x, y.method, y.args) if type(y) is ast.MethodCall else
    ast.Index(x, y.item) if type(y) is ast.Index else
    None, t[1:], t[0])
)

# var ::=  Name | prefixexp ‘[’ exp ‘]’ | prefixexp ‘.’ Name
# var <<= Name.copy().set_parse_action(lambda t: ast.Name(t[0])) | \
#   (prefixexp + LBRACK + exp + RBRACK).set_parse_action(lambda t: ast.Index(t[0], t[1])) | \
#    (prefixexp + PERIOD + Name).set_parse_action(lambda t: ast.Index(t[0], ast.LiteralString(t[1])))


def checkVar(t):
    if not isinstance(t[0], (ast.Index, ast.Name)):
        raise pp.ParseException(f"Expected ast.Index or ast.Name call, got {type(t[0])}")


var <<= prefixexp.copy().add_parse_action(checkVar)

# functioncall: := prefixexp args | prefixexp ‘:’ Name args


def checkFunctionCall(t):
    if not isinstance(t[0], (ast.Call, ast.MethodCall)):
        raise pp.ParseException(f"Expected call or method call, got {type(t[0])}")


functioncall <<= prefixexp.copy().add_parse_action(checkFunctionCall)

# functiondef: := function funcbody
functiondef <<= FUNCTION + funcbody

# funcbody: := ‘(’ [parlist] ‘)’ block end
funcbody <<= LPAR + pp.Optional(parlist, default=[]) + RPAR + block + END
funcbody.set_parse_action(lambda toks: ast.Func(args=toks[0], body=toks[1]))

# parlist: := namelist[‘,’ ‘...’] | ‘...’
parlist <<= pp.Group(namelist + (COMMA + ellipsis)[0, 1] | ellipsis, aslist=True)


# tableconstructor: := ‘{’ [fieldlist] ‘}’
tableconstructor <<= pp.Group(LBRACE - fieldlist[0, 1] - RBRACE, aslist=True)
tableconstructor.set_parse_action(lambda t: ast.Table(fields=t[0]))

# fieldlist: := field {fieldsep field}[fieldsep]
fieldlist <<= field + (fieldsep + field)[0, ...] + fieldsep[0, 1]

# field: := ‘[’ exp ‘]’ ‘=’ exp | Name ‘=’ exp | exp
field <<= (LBRACK + exp + RBRACK + EQ + exp).set_parse_action(lambda t: ast.Field(key=t[0], value=t[1])) | \
    (Name + EQ + exp).set_parse_action(lambda t: ast.Field(key=ast.LiteralString(t[0]), value=t[1])) | \
    (exp.copy()).set_parse_action(lambda t: ast.Field(value=t[0]))

# fieldsep ::= ‘,’ | ‘;’
fieldsep <<= COMMA | SEMI

# ignore comments, WARNING: has to be last, as it updates all the rules recursively
comment_intro = pp.Literal("--")
short_comment = pp.Regex(r"--(?![\|\{\}])") + pp.restOfLine
long_comment = pp.Regex(r"--\[(=*)\[(?P<str>[\s\S]*?)\]\1\]")
debug_comment = pp.Regex(r"--!(=*)\[[\s\S]*?--!\1\]")
lua_comment = debug_comment | long_comment | short_comment
block.ignore(lua_comment)


def parse_string(x):
    return chunk.parse_string(x, parse_all=True)[0]



from cmath import inf
import argparse
from pakettic import ast
from glob import glob
import os
import sys
import zlib
import pkg_resources
import zopfli
import time
import datetime

from pakettic import parser, printer, optimize, ticfile
import tqdm


def _parse_chunks_arg(arg: str) -> list[ticfile.ChunkID]:
    """Parse the --chunks command line arugment into a list of chunk types"""
    a = arg.upper()
    if a == 'ALL' or a == 'ALL_EXCEPT_DEFAULT':
        chunk_types = [e for e in ticfile.ChunkID if e !=
                       ticfile.ChunkID.CODE_ZIP and e != ticfile.ChunkID.DEFAULT]
        if a == 'ALL':
            chunk_types = [ticfile.ChunkID.DEFAULT] + chunk_types
    else:
        chunk_types = []
        for chunk_id_name in a.split(','):
            x = getattr(ticfile.ChunkID, chunk_id_name.upper())
            if x is None:
                sys.exit(f"Unknown chunk type entered on command line: {chunk_id_name}")
            chunk_types.append(x)
        chunk_types = list(dict.fromkeys(chunk_types))  # remove duplicates
    return chunk_types


_ZOPFLI_LEVELS = [
    {"iterations": 1, "split": False, "split_max": 0},
    {"iterations": 5, "split": False, "split_max": 0},
    {"iterations": 15, "split": True, "split_max": 1},
    {"iterations": 15, "split": True, "split_max": 3},
    {"iterations": 200, "split": True, "split_max": 0},
    {"iterations": 500, "split": True, "split_max": 0}
]


def _parse_zopfli_level(arg: str) -> dict:
    """Parse the --zopfli-level command line argument into a dictionary containing the zopfli parameters"""
    try:
        level = int(arg)
        if level < 0 or level > 5:
            raise argparse.ArgumentTypeError()
        return _ZOPFLI_LEVELS[level]
    except:
        raise argparse.ArgumentTypeError("Compression level should be an integer 0-5")


def _compress(bytes, split, split_max, iterations):
    """Compress a byte string using Zopfli, dropping the check sum"""
    c = zopfli.ZopfliCompressor(zopfli.ZOPFLI_FORMAT_ZLIB, block_splitting=split,
                                block_splitting_max=split_max, iterations=iterations)
    return (c.compress(bytes) + c.flush())[: -4]


def main():
    """Main entrypoint for the command line program"""
    sys.setrecursionlimit(100000)  # TODO: find out why the parser recurses so heavily and reduce that

    version = pkg_resources.get_distribution('pakettic').version
    argparser = argparse.ArgumentParser(
        prog='pakettic', description=f'Minify and compress TIC-80 fantasy console carts. v{version}',
        formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=33))
    argparser.add_argument('input', nargs='+', help='input file(s). ?, * and ** wildcards work')
    argparser.add_argument('-o', '--output', default=os.getcwd(), metavar='str', help='output file or directory')
    argparser.add_argument('-l', '--lua', action='store_const', const=True,
                           help='output .lua carts instead of .tic carts')
    argparser.add_argument('-u', '--uncompressed', action='store_const', const=True,
                           help='leave code chunks uncompressed, even when outputting a .tic file')
    argparser.add_argument('-p', '--print-best', action='store_const', const=True,
                           help='pretty-print the best solution when found')
    argparser.add_argument('-c', '--chunks',
                           default='code,default', metavar='str', help='chunk types to include and their order. valid: ALL, ALL_EXCEPT_DEFAULT, or comma-separated list without spaces: BINARY,CODE,COVER_DEP,DEFAULT,FLAGS,MAP,MUSIC,PALETTE,PATTERNS_DEP,PATTERNS,SAMPLES,SCREEN,SPRITES,TILES,WAVEFORM. default: %(default)s',
                           type=_parse_chunks_arg)
    argparser.add_argument('-d', '--data-to-code', action='store_const', const=True, default=False, help='convert data chunks into code, so they can be compressed')
    argparser.add_argument('--pedantic', action='store_const', const=True, default=False,
                           help='write DEFAULT chunk in full even when it is the last chunk')
    argparser.add_argument('--no-load', action='store_const', const=True, default=False,
                           help="disable converting function()end into load''")
    optgroup = argparser.add_argument_group('optional arguments for the optimization algorithm')
    optgroup.add_argument('-a', '--algorithm',
                          action='store',
                          type=str.lower,
                          metavar='str',
                          help="optimization algorithm, LAHC (late acceptance hill climbing), DLAS (diversified late acceptance search) or ANNEAL (simulated annealing). Default: %(default)s",
                          required=False,
                          choices=["lahc", "dlas", "anneal"],
                          default="dlas")
    optgroup.add_argument('-s', '--steps', type=int, default=10000, metavar='int',
                          help='number of steps in the optimization algorithm. default: %(default)d')
    optgroup.add_argument('-H', '--lahc-history', type=int, default=500, metavar='int',
                          help='history length in late acceptance hill climbing. default: %(default)d')
    optgroup.add_argument('-D', '--dlas-history', type=int, default=5, metavar='int',
                          help='history length in diversified late acceptance search. default: %(default)d')
    optgroup.add_argument('-m', '--margin', type=float, default=0, metavar='float',
                          help='initialize the lahc/dlas history with initial_cost + margin (in bytes). Default: %(default).0f')
    optgroup.add_argument('-t', '--start-temp', type=float, default=1, metavar='float',
                          help='starting temperature for simulated annealing, >0. default: %(default).1f')
    optgroup.add_argument('-T', '--end-temp', type=float, default=0.1, metavar='float',
                          help='ending temperature for simulated annealing, >0. default: %(default).1f')
    optgroup.add_argument('--target-size', type=int, default=0, metavar='int',
                          help='stop compression when target size is reached. default: %(default)d')
    optgroup.add_argument('--exact', action='store_const', const=True,
                          help='used with --target-size to indicate that the size should be reached exactly')
    optgroup.add_argument('--seed', type=int, default=0, metavar='int',
                          help='random seed. default: %(default)d')
    zopfligroup = argparser.add_argument_group('optional arguments for tuning zopfli')
    zopfligroup.add_argument('-z', '--zopfli-level', type=_parse_zopfli_level, default=_ZOPFLI_LEVELS[0], metavar='int',
                             help='generic compression level for zopfli, 0-5. default: 0')
    zopfligroup.add_argument('--iterations', type=int, metavar='int',
                             help='number of iterations in zopfli. default: based on compression level')
    zopfligroup.add_argument('--split', action=argparse.BooleanOptionalAction,
                             help='enable or disable block splitting in zopfli. default: based on compression level')
    zopfligroup.add_argument('--split-max', type=int, metavar='int',
                             help='maximum number of block splittings in zopfli (0: infinite). default: based on compression level')
    args = argparser.parse_args()

    if args.lua:
        args.uncompressed = True  # Outputting LUA and compressing are mutually exclusive
    if args.split is None:
        args.split = args.zopfli_level["split"]
    if args.split_max is None:
        args.split_max = args.zopfli_level["split_max"]
    if args.iterations is None:
        args.iterations = args.zopfli_level["iterations"]

    input = []
    # use glob to find files matching wildcards
    # if a string does not contain a wildcard, glob will return it as is.
    for arg in args.input:
        input += glob(arg, recursive=True)
    if len(input) > 1 and not os.path.isdir(args.output):
        sys.exit('When multiple input files are defined, the output must be a directory.')
    if len(input) == 0:
        sys.exit('No input files found.')
    input = sorted(input)  # sort the input files so corpus will be reported in same order
    filepbar = tqdm.tqdm(input, leave=False, smoothing=0.02)
    error = False
    total_original_size = 0
    total_minified_size = 0
    total_optimized_size = 0
    total_start_time = time.time()
    maxpathlen = max(len(p) for p in input)
    for filepath in filepbar:
        cart_start_time = time.time()
        _, filename = os.path.split(os.path.splitext(filepath)[0])
        ext = '.lua' if args.lua else '.tic'
        outfile = os.path.join(args.output, filename + '.packed' + ext) if os.path.isdir(args.output) else args.output
        original_size = os.path.getsize(filepath)
        filepath_sliced = filepath[-30:] if len(filepath) > 30 else filepath
        filepbar.set_description(f"Reading       {filepath_sliced}")
        try:
            cart = ticfile.join_code(ticfile.read(filepath))
        except Exception as e:
            filepbar.write(f"Error reading {filepath}: {e}, skipping...")
            error = True
            continue
        filepbar.set_description(f"Decompressing {filepath_sliced}")
        cart.update([((k[0], ticfile.ChunkID.CODE), zlib.decompress(v[2:], -15))
                    for k, v in cart.items() if k[1] == ticfile.ChunkID.CODE_ZIP])  # decompress zipped chunks
        cart[(0, ticfile.ChunkID.DEFAULT)] = b''  # add default chunk if it's missing
        cart = dict(c for i in args.chunks for c in cart.items() if c[0][1] == i)  # only include the chunks listed in args
        if args.data_to_code:
            ticfile.data_to_code(cart)
        filepbar.set_description(f"Compressing   {filepath_sliced}")
        outcart = cart.copy() if args.uncompressed else dict((k, v) if k[1] != ticfile.ChunkID.CODE else (
            (k[0], ticfile.ChunkID.CODE_ZIP), _compress(v, args.split, args.split_max, args.iterations)) for k, v in cart.items())
        final_size = ticfile.write(ticfile.split_code(outcart), outfile)
        code_chunks = [c for c in cart if c[1] == ticfile.ChunkID.CODE]
        for c in code_chunks:
            code = cart[c].decode("ascii")
            root = parser.parse_string(code)
            root = optimize.loads_to_funcs(root)
            root = optimize.minify(root)
            root = ast.Hint(root)

            def _cost_func(root, best_cost):
                nonlocal final_size
                bytes = printer.format(root, no_load=args.no_load).encode("ascii")
                key = c
                if not args.uncompressed:
                    key = (c[0], ticfile.ChunkID.CODE_ZIP)
                    bytes = _compress(bytes, args.split, args.split_max, args.iterations)
                diff = len(bytes) - len(outcart[key])
                ret = final_size + diff - args.target_size
                if args.exact:
                    ret = abs(ret)
                if ret < best_cost:
                    outcart[key] = bytes
                    final_size = ticfile.write(ticfile.split_code(outcart), outfile, pedantic=args.pedantic)
                    if args.print_best:
                        filepbar.write(f"-- {ret} bytes:\n{'-'*40}\n{printer.format(root, pretty=True).strip()}\n{'-'*40}")
                return ret
            _cost_func(root, inf)
            minified_size = final_size  # current final_size is the size after minification
            if args.algorithm == 'lahc':
                root = optimize.lahc(root, steps=args.steps, cost_func=_cost_func,
                                     list_length=args.lahc_history, init_margin=args.margin, seed=args.seed)
            elif args.algorithm == 'dlas':
                root = optimize.dlas(root, steps=args.steps, cost_func=_cost_func,
                                     list_length=args.dlas_history, init_margin=args.margin, seed=args.seed)
            else:
                root = optimize.anneal(root, steps=args.steps, cost_func=_cost_func,
                                       start_temp=args.start_temp, end_temp=args.end_temp, seed=args.seed)
        total_original_size += original_size
        total_optimized_size += final_size
        total_minified_size += minified_size
        cart_time_str = '.'.join(str(datetime.timedelta(seconds=int(time.time() - cart_start_time))).split(':'))
        filepbar.write(f"{filepath.ljust(maxpathlen)} Time:{cart_time_str} Orig:{original_size:<5} Min:{minified_size:<5} Pack:{final_size:<5}")
    if len(input) > 1:
        total_time_str = str(datetime.timedelta(seconds=int(time.time() - total_start_time)))
        print("-" * 80 + f"\n{'Totals'.ljust(maxpathlen)} Time:{total_time_str} Orig:{total_original_size:<5} Min:{total_minified_size:<5} Pack:{total_optimized_size:<5}")
    sys.exit(1 if error else 0)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pakettic']

package_data = \
{'': ['*']}

install_requires = \
['pyparsing>=3.0.8,<4.0.0', 'tqdm>=4.64.0,<5.0.0', 'zopflipy>=1.7,<2.0']

entry_points = \
{'console_scripts': ['pakettic = pakettic.main:main']}

setup_kwargs = {
    'name': 'pakettic',
    'version': '1.3.0',
    'description': 'A tool for minifying and compressing TIC-80 fantasy console carts',
    'long_description': '# Pakettic\n\nPakettic is a command-line tool for minifying and compressing\n[TIC-80](http://tic80.com/) fantasy console carts. The tool is written\nin Python (3.9+) and used especially for\n[sizecoding](http://www.sizecoding.org/wiki/TIC-80). It compresses\nexisting carts approximately ~1.2% better than best alternatives, and by\nusing its [magic comments](#magic-comments), pakettic might find code\nthat compresses even better.\n\n## Installation\n\nInstalling with pip:\n\n```bash\n$ pip install pakettic\n```\n\nInstalling the latest main branch from GitHub:\n\n```bash\n$ pip install git+https://github.com/vsariola/pakettic.git@main\n```\n\nInstalling a checked out version of the repository:\n\n```bash\n$ pip install -e path/to/pakettic\n```\n\nInstalling a checked out version of the repository using\n[poetry](https://python-poetry.org/) for a nice virtual environment with\nlocked dependencies (run inside the pakettic folder):\n\n```bash\n$ poetry install\n```\n\n## Usage\n\nTo compress a cart, run:\n\n```bash\n$ pakettic path/to/cart.tic\n```\n\nIf your PATH is not configured to include pip installed executables, you\ncan use\n\n```bash\n$ python -m pakettic path/to/cart.tic\n```\n\nIf you installed using poetry into a virtual environment, you need to\nprepend `poetry run` before every command e.g.\n\n```bash\n$ poetry run pakettic path/to/cart.tic\n```\n\nPakettic supports both .tic and .lua carts. Multiple input files may be\ndefined. Input files are globbed, so `?`, `*`, and `**` work as\nwildcards for a single character, multiple characters and a directory,\nrespectively.\n\nFor a full list of command line options, see:\n\n```bash\n$ pakettic --help\n```\n\nSee also [tips for command line arguments](#tips-for-command-line-arguments)\n\nRunning all tests:\n\n```bash\n$ poetry run python -m unittest discover -s tests\n```\n\n## Features\n\nPakettic first parses the LUA-script to an abstract syntax tree, and\nthen uses a local optimization algorithm\n([simulated annealing](https://en.wikipedia.org/wiki/Simulated_annealing),\n[late acceptance hill climbing](https://arxiv.org/pdf/1806.09328.pdf) or\nits variant diversified late acceptance search) to randomly mutate the\nsyntax tree & see if it compresses better. Implemented mutations\ninclude:\n  - shortening variable names\n  - flipping comparisons `>`, `<`, `>=`, `<=`, `~=`, and `==`\n  - reordering arithmetic operators `+`, `-`, `*` and `/` and bit logic\n    operators `&`, `~` and `|`\n  - converting `a^2` into `a*a` and vice versa\n  - using either single or double quotes for all strings\n  - converting whole hexadecimals into decimals\n  - convert `for a,b,1 do` into `for a,b do` and vice versa\n  - reordering statements: statements that can be reordered are marked\n    with [magic comments](#magic-comments)\n  - alternative expressions: alternatives are marked with\n    [magic comments](#magic-comments)\n  - folding constant expressions\n\nInternally, pakettic uses [zopfli](https://github.com/google/zopfli) for\nthe compression.\n\n`load\'<code>\'` is parsed as `function(...)<code>end` so you can easily\nrecompress already compressed carts. Conversely, `function()<code>end`\nor `function(...)<code>end` is replaced with `load\'<code>\'` during\ncompression.\n\nNote that `function(...)<code>end` and `load\'<code>\'` are not 100%\nsemantically identical: the load version cannot access locals in the outer\nscope. For example: `local x="hello" function f()print(x)end` works but\n`local x="hello" f=load\'print(x)\'` does not. Since locals are rarely\nused in size-coding, we default to using the load-trick, but you can\ndisable it with the command-line parameter `--no-load`.\n\nHowever, pakettic does not convert functions with parameters. In\nparticular, pakettic does not automatically convert\n`function SCN(x)<code>end` into `SCN=load\'x=...<code>\'`, because they\nare not semantically identical: in the load version, `x` is now global\nand thus could trash a global variable, unintentionally breaking the\ncart. To make `SCN` compress nicely, you have to write it as\n`function SCN(...)x=...<code>end`, taking responsibility for `x` not\noverwriting anything important.\n\nUnnecessary parentheses are removed from expressions so you do not have\nto worry about those.\n\n## Magic comments\n\n### Reorderable statements\n\nThe algorithm will try to reorder statements between `--{` and `--}`.\nFor example:\n\n```lua\n --{\n a="hello"\n b="world"\n --}\n```\n\nwill try both `a="hello"b="world"` and `b="world"a="hello"` to see which\ncompresses better.\n\nNotice that only complete statements can be reordered. Thus, this will\nNOT work:\n\n```lua\n --{\n for x=0,239 do\n  for y=0,135 do\n --}\n  end\n end\n```\n\nA good rule of thumb is that you should be able to replace `--{` and\n`--}` with `do` and `end`, respectively, and still have valid code.\n\nStatements between `--{!` and `--}` are not ordered, so you can make\nblocks of statements that are kept in order within a pair of `--{` and\n`--}` tags.\n\n### Alternative expressions\n\nThere is a special `--|` operator that allows alternative expressions to\nbe tested, to see if they compress better. For example: `5--|4--|6`\nmeans that the algorithm will try 4 and 6 in place of the 5. This will\nnaturally show up as a comment in LUA so you will have to continue the\nexpression on next line if this is in the middle of an expression. `--|`\nhas the lowest precedence, even lower than `^`, so put parentheses if\nyou want to try more complicated expressions e.g. `(x//256)--|(x>>8)`\n\n### Debug code\n\nPakettic treats `--![` and `--!]` as multiline comment tags, while LUA\ntreats these as single line comments. Useful for including debug code in\nthe unpacked intro: the code will not be included in the packed cart.\n\n## Tips for command line arguments\n\n- If pakettic complains about CODE_ZIP chunk size, the code is just too\n  big after compression. In TIC-80, CODE_ZIP chunks do not support\n  multiple banks (and likely never will, as the feature is already\n  deprecated), and thus are unfortunately limited to 65535 bytes.\n  `--uncompressed` is a temporary fix, but code will be uncompressed and\n  thus the size much larger.\n- The Zopfli compression level can be set with `-z<level>`, with level\n  ranging from 0 to 5. When developing, start with `-z0` for fast\n  optimization, and only increase when necessary e.g. when you are just\n  a few bytes over the limit. The default Zopfli-level is 0.\n- The algorithm uses a pseudorandom generator. Sometimes using a\n  different seed finds a few byte better or worse solution. Use command\n  line argument `--seed` to try different seeds.\n- Similarly, different optimization heuristics produce slightly\n  different results. Try different heuristics e.g. with `-alahc`,\n  `-adlas` or `-aanneal`.\n- To avoid re-optimizing all the expressions every time, do a long\n  optimization run, study the results and change your expressions to the\n  forms that pack well. Set the number of steps with `-s`. Use\n  command-line argument `-p` to always print a reasonably readable\n  version of the best solution when one is found.\n- By default, pakettic only includes CODE and DEFAULT chunks. DEFAULT\n  indicates that before loading the cart, TIC-80 loads the default cart,\n  setting default palette, waveforms etc. If you don\'t need the default\n  values (e.g. you set the palette yourself), save one byte by only\n  including CODE chunk in the cart: `-ccode`\n- Working on a tweet-cart? Use `-l` to output LUA carts, which are\n  uncompressed. The optimization algorithm then just optimizes the\n  uncompressed size of the code.\n- If the packed cart is missing sprites, music, map etc., try adding\n  `-call` (or something more specific) to include necessary chunks.\n- Do you want to use the TIC-80 sprites or the tracker, but don\'t like\n  the fact that the data chunks are uncompressed? Use `-d` to have\n  pakettic automatically convert all data chunks into hexadecimal\n  strings in the code, along with a small stub placed at the beginning\n  of the code that interprets the string and loads the data at correct\n  address. For example,\n  `-d -cCODE,MUSIC,PATTERNS,WAVEFORM,SAMPLES,DEFAULT` would include the\n  necessary chunks for the music.\n\n## Known issues\n\n- At the moment, all the branches of swappable operators are assumed to\n  be without side effects. If they have side-effects, the swapping might\n  inadvertedly swap the execution order of the two branches.\n- The parser can crash with large carts. Carts in the size coding range\n  (few thousand characters) do not seem to cause problems, but crashes\n  have been observed parsing carts with tens of thousands of code\n  characters. This may be related to how the pyparsing grammar is\n  defined, which could result in highly recursive parsing and eventually\n  stack overflows.\n\n## Credits\n\nCode contributors: [Veikko Sariola/pestis](https://github.com/vsariola), [wojciech-graj](https://github.com/wojciech-graj),\n[koorogi](https://github.com/koorogi)\n\nTest corpus contributors: [psenough](corpus/psenough/), [ilmenit](corpus/ilmenit/),\n[gigabates](corpus/gigabates/), [gasman](corpus/gasman/), [pellicus](corpus/pellicus/),\n[luchak](corpus/psenough/fabracid.lua).\n\n## License\n\n[MIT](https://choosealicense.com/licenses/mit/)\n\nThe test corpus carts have their own licenses, see the license files in\nthe subdirectories of the [corpus](corpus/) directory.\n',
    'author': 'Veikko Sariola',
    'author_email': '5684185+vsariola@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/vsariola/pakettic',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
